
#
# Trax (network activity tracker) Version 1.3
#
# Copyright 2023 Brandyn Webb
# 
# This file is part of Trax (network activity tracker).
# 
# Trax is free software: you can redistribute it and/or modify it under the terms 
# of the GNU General Public License as published by the Free Software Foundation, 
# either version 3 of the License, or (at your option) any later version.
# 
# Trax is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with 
# Trax. If not, see <https://www.gnu.org/licenses/>. 
# 

from datetime import datetime
import traceback

debug_gc = True

if debug_gc:
    from weakref import WeakValueDictionary

class Logger(object):

    """A Logger is a superclass for objects that want to log or send messages
        to one or more *globally defined* channels.  Mainly its purpose is to
        standardize the formatting and centralize the routing of common messages.

    In the typical case, most of these messages will originate as self.log() calls,
        and so this module just defines an easy way to re-route routine logging
        output when desired.

    But this also provides for general, configurable routing of calls with
        arbitrary function signatures (provided there is a well defined signature
        for each declared channel name).

    I.e., mostly this is just a global (and globally configurable) dictionary
        mapping channel names to handler functions, with some wrappers on top
        which use this for generic, configurable logging and alerts.

    Pre-defined channels, which you can re-direct (with cls.log_channel_handler)
        as long as you obey the same calling conventions:

    log_to('raw-log', msg)                                           - writes msg to stdout.  Redirect this if you just want to change where typical log message go.
    log_to('admin', msg)                                             - sends msg to a human admin if possible.  (Default just logs it.)
    log_to('log', sender_name, msg, err=None, show_traceback=True)   - Formats msg as from sender with current timestamp, and optional error, and sends to 'raw-log' channel.
    log_to('alert', sender_name, msg, err=None, show_traceback=True) - Adds flare to msg, then sends to 'log' channel.  Suggest redirecting this to direct message to admin.

    You can override these, or add custom channels, with log_channel_handler

    I *think* it should be possible to subclass Logger into multiple subclasses
        which have their own log_channels, if you want to have more than one
        group of channels under the same name.  (I.e., if you wanted to have
        some objects logging one place and others another, you could subclass
        the two sets from different subclasses of Logger and then set their
        respective raw-log handlers accordingly.)
    """
    __slots__ = ('log_name', '__weakref__')

    log_channels = {}   # Map from channel name to handler function.

    if debug_gc:
        num_active = 0
        gc_active  = WeakValueDictionary()

    @classmethod
    def log_channel_handler(cls, channel, handler):
        """This sets the raw handler function for a given channel.
        Here it is assumed that handler obeys the implicit calling
        conventions of the channel.
        """
        cls.log_channels[channel] = handler

    @classmethod
    def log_channel_disable(cls, channel):
        """This sets the raw handler function for a given channel
        to /dev/null.
        """
        cls.log_channel_handler(channel, null_handler)

    def __init__(self, log_name=None):
        """Log_name is the name we sign our messages with (to all channels).
        """
        if log_name is None:
            log_name = "%s.%d"%(self.__class__.__name__,id(self)%10000)

        self.log_name = log_name

        if debug_gc:
            Logger.gc_active[log_name] = self
            Logger.num_active += 1

    def log(self, msg, err=None, show_traceback=True):
        """This is the most common function to call.  Typically as

        self.log("Some message", e)

        Where e is an optional Exception that just happened.

        It will log the provided message (and error if applicable) with
            a timestamp and this object's log_name to whatever the configured
            log output is (stdout by default).
        """
        self.log_to('log', self.log_name, msg, err=err, show_traceback=show_traceback)

    def warn(self, msg, err=None, show_traceback=True):
        """Just like log() but slightly higher priority notification
            and prepends "WARNING:" or similar.

        Where 'log' entries may just be routine status, 'warn' should
            things that might warrent attention, but which are common
            or non-critical.  Typically these might go to a low-traffic
            log for regular review, but do not demand immediate attention.

        By default warns go the same place as logs; but they are here
            so it's easy to separate out this channel when desired.
        """
        self.log_to('warn', self.log_name, f"WARNING: {msg}", err, show_traceback)

    def alert(self, msg, err=None, show_traceback=True):
        """Just like log() but higher priority notification.

        Typically this would be configured to send a message
            directly to an admin, in addition to logging it.
        """
        self.log_to('alert', self.log_name, msg, err, show_traceback)

    def notify_admin(self, msg):
        """This logs message, and then sends it directly to a human admin when possible.
        """
        self.log_to('admin', msg)

    @classmethod
    def log_to(cls, channel, *args, **kwargs):
        """This is the primary call-to-channel dispatcher, which is mostly
            trivial except for some error handling.

        The channel is specified as a simple name, commonly 'log', 'warn' and 'raw-log',
            sometimes 'alert', and in some applications any number of other
            (globally) registered channels.

        If a message is sent to an unknown channel, this fact will be
            logged (once) via an alert, and then that channel will be
            redirected to the standard log.
        """
        if channel not in cls.log_channels:
            cls.log_to('alert', cls.__name__, "Tried to log to channel %r but no such exists."%(channel,))
            def logfunc(*args, **kwargs):
                cls.log_channel_unknown(channel)
            cls.log_channels[channel] = logfunc

        cls.log_channels[channel](*args, **kwargs)

    @classmethod
    def log_channel_rawlog(cls, msg):
        """This is the default handler for the "raw-log" channel,
            which just prints a message to stdout.
        """
        print(msg, flush=True)

    @classmethod
    def log_channel_log(cls, sender_name, msg, err=None, show_traceback=True):
        """This is the default handler for the "log" channel,
            which formats the message and any error and forwards
            it to the raw-log channel.
        """
        cls.log_to('raw-log', cls.log_format(sender_name, msg, err, show_traceback))

    @classmethod
    def log_channel_alert(cls, sender_name, msg, err=None, show_traceback=True):
        """This is the default handler for the "alert" channel,
            which just adds some flourish to the message and forwards
            it to the log channel.
        """
        msg = "!!!ALERT!!!  " + msg
        cls.log_to('log', sender_name, msg, err, show_traceback)
        cls.log_to('admin', "%s: %s"%(sender_name, msg))

    @classmethod
    def log_channel_admin(cls, msg):
        """This is the default handler for the "admin" channel,
            which for now just logs the message but this should
            be overriden to actually send the message to an admin.
        """
        msg = "TO ADMIN: " + msg
        cls.log_to('log', 'Logger', msg)

    @classmethod
    def log_channel_unknown(cls, channel):
        """Helper function for messages sent to unregistered channels.
        """
        if False:   # These could get out of hand, so we'll just no-op it.
            cls.log_to('log', "Logger", "<unhandled message to %r>"%(channel,))

    @classmethod
    def log_format(cls, sender_name, msg, err=None, show_traceback=True):
        """This takes a message and optional error and formats it into
            a string in a standard way with the current timestamp and such.

        Note the returned string may have embedded carriage returns, but
            will not end with one.
        """
        header = "%s %s: "%(datetime.now(), sender_name)
        r = header + msg
        if err:
            r += '\n'+cls.log_err_format(err, show_traceback)
        return cls.log_indent(r, " "*len(header))

    @classmethod
    def log_err_format(cls, err, show_traceback=True):
        """This simple utility just formats the provided error as
            a multi-line string using the *current traceback* (unless
            show_traceback is False).  In other words, this may
            be bogus if err is taken from too far afield.
        """
        if show_traceback:
            s = traceback.format_exc()
            if s:
                return "---- ERROR: %s\n%s"%(err, s.strip())
        return "---- ERROR: %s"%(err)

    @classmethod
    def log_indent(cls, msg, indent):
        """This simple utility indents a multi-line string by the given indent string,
            but only after embedded returns (so you need to add indent to the
            begining if you want the first line indented as well).

        If msg is a single line only, this has no effect.
        """
        return msg.replace("\n", "\n"+indent)

    def __str__(self):
        """Fine for subclass to override -- just handy default.
        Not counted on anywhere.
        """
        return self.log_name

    if debug_gc:
        @staticmethod
        def debug_gc_dump(full=False):
            print(f"Active Loggers: {Logger.num_active}")
            print(f"Uniquely Named: {len(Logger.gc_active)}")
            if full:
                for name, l in Logger.gc_active.items():
                    print(f"  {name}")

        def debug_gc_format(full=False):
            """Returns a list of lines."""
            lines = []
            lines.append(f"Active Loggers: {Logger.num_active}")
            lines.append(f"Uniquely Named: {len(Logger.gc_active)}")
            if full:
                for name, l in Logger.gc_active.items():
                    lines.append(f"  {name}")
            return lines

        def __del__(self):
            Logger.num_active -= 1
            # object doesn't have a __del__ to call.

def null_handler(*args, **kargs):
    pass

Logger.log_channel_handler('raw-log', Logger.log_channel_rawlog)
Logger.log_channel_handler(    'log', Logger.log_channel_log)
Logger.log_channel_handler(   'warn', Logger.log_channel_log)
Logger.log_channel_handler(  'alert', Logger.log_channel_alert)
Logger.log_channel_handler(  'admin', Logger.log_channel_admin)



if __name__ == "__main__":

    class Foobar(Logger):
        pass

    foo = Foobar()

    foo.log("Ho de do.")
    try:
        x = 1/0
    except Exception as e:
        foo.log("testing...\nTesting...\n1..2..3..", e)
    foo.log("La de da.")
    foo.alert("This is an alert!")
    foo.notify_admin("This is a notification to the admin.")
    foo.log_to('bogus', "Doesn't matter what's here...")
    foo.log_to('bogus', "Doesn't matter what's here...")
    foo.log_to('bogus', "Doesn't matter what's here...")
    foo.log_to('bogus', "Doesn't matter what's here...")

