
#
# Trax (network activity tracker) Version 1.3
#
# Copyright 2023 Brandyn Webb
# 
# This file is part of Trax (network activity tracker).
# 
# Trax is free software: you can redistribute it and/or modify it under the terms 
# of the GNU General Public License as published by the Free Software Foundation, 
# either version 3 of the License, or (at your option) any later version.
# 
# Trax is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with 
# Trax. If not, see <https://www.gnu.org/licenses/>. 
# 

import trio

from Logger import Logger
from Server import on_shutdown
from TrioUtils import spawn, pool, waitfor

class Process(Logger):

    """The Process class mostly handles graceful shutdown of long-running entities.

    The main API provided is:

        run()          - async method to run the process.
        quit()         - regular method to initiate shutdown.
        shutdown()     - async shutdown.

    Also for some circumstances:

        pre_shutdown() - prepare for imminent shutdown, for coordinating with other fully-operational Processes.

    Subclasses provide a run_() method which is the main run loop, as well
        as optional pre_shutdown_() and shutdown_() methods which do any needed
        cleanup.  Any threads still running after shutdown_() exits will be
        Canceled (which is a fine default where no special cleanup is required).

    It's possible to re-run a quit Process (once it has fully stopped) if the subclass supports it.
    """
    def __init__(self, process_manager, log_name=None):
        """log_name is the canonical, global name of this particular process, like 'waveplus1'.

        Self is automatically added to process_manager.
        """
        Logger.__init__(self, log_name)

        self.process_manager = process_manager

        #
        # When a process is stopped (including after it is shut down) it will look like this:
        #
        self.process_state    = 'stopped'   # One of: 'stopped', 'running', 'pre_shutdown', 'shutting_down'
        self.process_pool     = None        # This will be the top level pool in which run_() is running, which can be used to spawn more threads if desired.
        self.process_quit     = None        # This will be an Event when process starts, and will be set() when shutdown commences.
        self.process_complete = None        # This will be an Event when process starts, and will be set() when process ends (then cleared to None).

        self.process_prequeue = []          # any awaitables spawned before run is called.

        process_manager += self

    async def run(self):
        """Starts the main run loop of the Process.

        Returns when the process is complete.

        If called when already running, will just wait until the process ends and then return.
        """
        if self.process_state == 'stopped':

            self.process_state    = 'running'
            self.process_quit     = trio.Event()
            self.process_complete = trio.Event()

            self.log("Starting up.")

            async with pool() as p:
                p.start_soon(self.run_wrapper)
                p.start_soon(self.shutdown_when_its_time)

            self.log("Shut down.")

            self.process_complete.set()

            self.process_state    = 'stopped'
            self.process_complete = None
            self.process_quit     = None

        else:
            await self.process_complete.wait()
            return

    def quit(self):
        """This immediately-returning function triggers graceful shutdown
        and will eventually result in run() returning.
        """
        if self.process_quit is not None:
            self.process_quit.set()

    async def shutdown(self):
        """Like quit() but doesn't return until the process is done.
        """
        if self.process_quit is not None:
            self.process_quit.set()
            await self.process_complete.wait()

    async def pre_shutdown(self):
        """A ProcessManager calls this method once on all processes before an orderly
            shutdown is commenced.  It is an opportunity to doing any tidying up which
            may require coordination with other fully operational processes.

        All pre_shutdowns are invoked in parallel, and the true shutdown phase
            begins once the last one completes.

        This method should only be called once.  Generally only by the ProcessManager.
        """
        if self.process_state == 'running':
            self.log("Initiating any pre-shutdown cleanup.")
            self.process_state = 'pre_shutdown'
            await self.pre_shutdown_()
            await trio.sleep(.01)                       # Give run_wrapper a chance to clean up.

    def spawn_os_process(self, procname, func, *args, **kargs):
        self.process_manager.spawn_os_process(procname, func, *args, **kargs)

    def await_os_process(self, procname, func, *args, **kargs):
        """This is async.  See process manager's version."""
        return self.process_manager.await_os_process(procname, func, *args, **kargs)

    def spawn(self, awaitable):
        if self.process_pool is None:
            self.process_prequeue.append(awaitable)
        else:
            spawn(self.process_pool, awaitable)

    #================ Don't call anything below here directly =================

    async def run_wrapper(self):
        """Wraps run_ in self.process_pool so we can Cancel it.
        Also starts any awaitables already spawned.
        """
        if self.process_quit.is_set():                  # In case we've been quit before we could even start.
            return

        async with pool() as self.process_pool:
            self.process_pool.start_soon(self.run_)
            while self.process_prequeue:
                spawn(self.process_pool, self.process_prequeue.pop())
        self.process_pool = None

        self.process_quit.set()                         # Necessary to terminate shutdown_when_its_time if run_ quits on its own.

    async def shutdown_when_its_time(self):

        #
        # Wait until it's time to shut down:
        #
        await self.process_quit.wait()

        #
        # If we haven't already entered pre_shutdown mode, go ahead and
        #  do it now (usually this would only happen if we are being
        #  individually shut down as opposed to when the whole ProcessManager
        #  is shutting down):
        #
        if self.process_pool and self.process_state == 'running':
            await self.pre_shutdown()

        if self.process_pool:
            self.log("Attempting graceful shutdown...")
            self.process_state = 'shutting_down'
            await self.shutdown_()
            await trio.sleep(.01)                       # Give run_wrapper a chance to clean up.

        if self.process_pool:
            self.log("Killing run loop.")
            self.process_pool.cancel_scope.cancel()

    #
    # === Subclasses should provide:
    #

    async def run_(self):
        """This long-running thread will be called once per run/shutdown cycle.
        It may exit any time, including in response to pre_shutdown_(), shutdown_(),
        or not at all.  If not at all, then it will be Canceled after shutdown_()
        completes.

        NOTE that any threads started with self.spawn() act as extensions of run_
            such that run_ is not really treated as done until they are too.
            Furthermore, exceptions in any spawned threads kill all spawned
            threads along with run...
        """
        return

    async def shutdown_(self):
        """This optional method will be called once per run/shutdown cycle to
            do any graceful last-minute tidying up.  Run_() may exit after this
            but that is not required -- any residual threads will be Canceled
            once shutdown_() returns.
        """
        return

    async def pre_shutdown_(self):
        """This is called once on all processes (in parallel) before shutdown_()
            is called, in order to give processes a chance to do any last minute
            tidying that might require coordination with other fully-functional
            processes.
        """
        return


class ProcessManager(Logger):

    """Master list of all Processes.  Mainly to manage shutdown cleanly.

    Process objects don't require a ProcessManager to run, so this is just
        an optional convenience.  (Well, it looks like I've made a process
        manager a parameter to the Process, but I could make that optional...)
    """
    def __init__(self, procs=[]):
        Logger.__init__(self, "ProcessManager")

        self.procs         = []             # List of all known processes.  (Order may matter.)
        self.proc_map      = {}             # Maps name to the corresponding process.
        self.top_pool      = None           # Top process pool in Trio, used to kill everything with Cancel.
        self.shutting_down = trio.Event()   # Set once we've been asked to shut everything down.

        # Rarely used, for handling OS-level multiprocessing (true fork()):
        self.os_logq  = None
        self.os_procs = []
        self.os_exits = {}

        on_shutdown(self.quit)              # Trap various signals to shut down gracefully.

        for proc in procs:
            self += proc

    def register_process(self, proc):
        """This adds a process to be run.

        If the ProcessManager is already running, the process will be started immediately.
        """
        name = proc.log_name

        if name in self.proc_map:
            self.alert("Process %r added more than once!"%(name,))
            return

        self.procs.append(proc)
        self.proc_map[name] = proc

        if self.top_pool is not None:   # The pool is already running, so start the process now.
            self.log("Launching %r on the fly."%(name,))
            self.top_pool.start_soon(proc.run)

        return proc

    def __iadd__(self, proc):
        """Add a process (will run it if we're already running)
        If proc is a list or tuple of processes, then they are all added.
        """
        if isinstance(proc, (tuple, list)):
            for p in proc:
                self += p
        else:
            self.register_process(proc)
        return self

    def get(self, devname):
        "Get process by name."
        return self.proc_map.get(devname)

    def __getitem__(self, devname):
        "Get process by name."
        return self.proc_map[devname]

    def __len__(self):
        "Number of processes"
        return len(self.procs)

    def __iter__(self):
        "Iterator over all processes."
        return iter(self.procs)

    def run(self):
        """Launch the main run loop using Trio.

        This will return when the last process exits (or is killed by quit).
        """
        trio.run(self.run_trio)

    def quit(self):
        """This immediately-returning function triggers graceful shutdown
        and will eventually result in run() returning.
        """
        self.shutting_down.set()

    #================ Don't call anything below here directly =================

    async def run_(self):
        """Subclass may use this for anything.
        This will be Canceled at the end of the shutdown process.
        """
        pass

    async def run_trio(self):
        async with pool() as p:
            p.start_soon(self.run_processes)
            p.start_soon(self.shutdown_when_its_time)

    async def run_processes(self):
        """Returns when the last process dies.
        """
        self.log("Launching %d processes."%(len(self),))
        async with pool() as self.top_pool:
            self.top_pool.start_soon(self.run_)
            for proc in self.procs:
                self.top_pool.start_soon(proc.run)
        self.top_pool = None    # All the processes have exited, so no need to ask them to shut down later.
        self.log("Last processes has exitted.")

        self.quit() # When our children are all done, so are we.

    async def shutdown_when_its_time(self):

        await self.shutting_down.wait()

        if self.top_pool:

            #
            # Notify all the processes we're going to shut them down:
            #
            self.log("Attemping graceful shutdown.")
            async with pool() as p:
                for proc in self.procs:
                    p.start_soon(proc.pre_shutdown)

        if self.top_pool:

            #
            # Notify all the processes to shut themselves down:
            #
            async with pool() as p:
                for proc in self.procs:
                    p.start_soon(proc.shutdown)

            await trio.sleep(.01)     # Give run_processes a chance to exit gracefully and clear top_pool before we Canel it.

        if self.top_pool:

            #
            # At this point just cancel any remaining threads, brute force.
            #
            self.log("Still some left.  Killing the rest.")
            self.top_pool.cancel_scope.cancel()

    #----------- Everything below here is just to support optional OS level multiprocessing (true process fork) ---------

    def spawn_os_process(self, procname, func, *args, **kargs):
        """This forks this process at the OS level and then calls
            func(*args, **kargs) in that new process.  Note func
            should be strictly synchronous -- all trio processing
            halts in the forked process.

        Logging will be set up in the new process to log via this
            parent process so that existing Logging output redirection
            is maintained.

        Returns the associated multiprocessing.Process object.
        """
        self.log(f"Forking OS process {procname}")
        import multiprocessing
        if self.os_logq is None:
            self.os_logq = multiprocessing.Queue()
        mproc = multiprocessing.Process(target=self.launch_os_process, args=(func, args, kargs, procname), name=procname)
        if not self.os_procs:
            self.top_pool.start_soon(self.track_os_procs)
        self.os_procs.append(mproc)
        mproc.start()
        return mproc

    async def await_os_process(self, procname, func, *args, **kargs):
        """Just like spawn, but this waits for the process to exit
            and then returns its exitcode.  This polls rapidly while
            the process is running, so this isn't suitable for long-
            running processes, rather for medium weight tasks that
            need to be backgrounded to avoid locking the event loop,
            but which can't be run in a lightweight process thread
            for whatever reason.

        This may take up to a second to notice the process has finished.
        """
        from TrioUtils import Event3

        mproc = self.spawn_os_process(procname, func, *args, **kargs)
        e     = Event3()
        self.os_exits[mproc] = e    # This will get set to the exit code by track_os_procs
        return await e.wait()

    #---- Don't call anything below here ----

    def launch_os_process(self, func, args, kargs, procname):
        Logger.log_channel_handler('raw-log', self.os_logq.put)
        self.log(f"{procname} forked.")
        func(*args, **kargs)

    def track_os_queues(self, inq, outq):
        """This runs in a background thread, transferring items from
            the inter-process queue to the inter-thread queue...

        You can kill it by putting None on the os_logq
        """
        while True:
            msg = inq.get()
            if msg is None:
                break
            outq.push_from_thread(msg)

    async def track_os_procs(self):
        """All of this is really just to "join" spawned Processes
            when they exit, so that we don't leave zombie processes around.
        """
        from TrioUtils import Queue, spawn

        self.log("Starting process tracker.")

        logq = Queue(threaded=True)
        spawn(self.top_pool, trio.to_thread.run_sync(self.track_os_queues, self.os_logq, logq, cancellable=True))

        while self.os_procs:
            for p in list(self.os_procs):  # Copy because we're editing the list as we go.
                if p.exitcode is not None:
                    self.log(f"Process {p.name} exitted.  Joining...")
                    if p in self.os_exits:               # If someone's waiting to be notified of the exit code, send it.
                        self.os_exits[p].set(p.exitcode)
                        del self.os_exits[p]
                    p.join()
                    p.close()
                    self.log(f"Done.")
                    self.os_procs.remove(p)
            with trio.move_on_after(1):      # Process the log queue for one second then re-check for dead procs.
                async for msg in logq:
                    self.log_to('raw-log', f">>{msg}")

        self.log("Stopping process tracker.")
        # FIXME: It has happened that we've gotten here with self.os_logq=None.  Figure out how that's possible and fix it...
        self.os_logq.put(None)  # This terminates track_os_queues now that it's no longer needed.
        self.os_logq = None     # This lets us know we need to restart that next time.

if __name__ == "__main__":

    import time

    pm = ProcessManager()

    class foo(Process):

        async def run_(self):

            self.spawn_os_process('multi_test', self.multi_test, 10, 20)

            self.log(f"In the main process..")
            for i in range(5):
                await trio.sleep(1)
                self.log("Dong.")

        def multi_test(self, x, y):
            self.log(f"In another process, we have x:{x} and y:{y}")
            for i in range(5):
                time.sleep(1)
                self.log("Ding.")

    foo(pm, "foo")

    pm.run()

