
#
# Trax (network activity tracker) Version 1.3
#
# Copyright 2023 Brandyn Webb
# 
# This file is part of Trax (network activity tracker).
# 
# Trax is free software: you can redistribute it and/or modify it under the terms 
# of the GNU General Public License as published by the Free Software Foundation, 
# either version 3 of the License, or (at your option) any later version.
# 
# Trax is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with 
# Trax. If not, see <https://www.gnu.org/licenses/>. 
# 

#
# Thin wrapper around curses just to simplify it slightly.
#
# This version of TextWin.py has been migrated from Io to trio.
#

import re   # Only used to scrub tabs and other funky chars when scrub=True
import os   # Just for setting the escape delay
import curses, sys, signal

from trio import open_nursery as pool, Event
from trio.lowlevel import wait_readable
from TrioUtils import CloseableQueue

class Window(object):

    # Attributes:
    Blink     = curses.A_BLINK      # Blinking text
    Bold      = curses.A_BOLD       # Extra bright or bold text
    Dim       = curses.A_DIM        # Half bright text
    Reverse   = curses.A_REVERSE    # Reverse-video text
    Standout  = curses.A_STANDOUT   # The best highlighting mode available
    Underline = curses.A_UNDERLINE

    # Colors (use colorPairAttribute to convert pairs of these into attributes you can acutally use):
    Black     = curses.COLOR_BLACK
    Blue      = curses.COLOR_BLUE
    Cyan      = curses.COLOR_CYAN
    Green     = curses.COLOR_GREEN
    Magenta   = curses.COLOR_MAGENTA
    Red       = curses.COLOR_RED
    White     = curses.COLOR_WHITE
    Yellow    = curses.COLOR_YELLOW

    weirdChars = re.compile('\s')   # Should improve this to be any non-printable, but don't know a quick way to do that offhand.

    #
    # .x,y          - Position in *screen* coordinates of inside of window (inside of border, if any).
    # .width,height - Size of inside of the window (does NOT include border if any).
    #
    def __init__(self, parent=None, x=None, y=None, width=None, height=None, border=False, color=False, scrub=False):
        """With no parent, returns the full screen.  color=True may be passed in that case to enable color.

        With a parent and window bounds, returns a sub-window of parent.

        Only the root window has these:

        .keystrokes - trio CloseableQueue input characters (ascii or special function code).
        .resizes    - trio CloseableQueue of (width, height) tuples each time we're resized (including initial size).
                      This size is internal to any border, and is the size of the whole root window.

        In either case:

        border    - if True, a border is drawn around the window and the coordinates are inset by 1

        scrub     - enable scrubbing of strings (replace weird characters with space).
                    (This is a convenience, but is less efficient, so only enable for dirty applications.)

        Note for the root window (full screen), the terminal mode is changed immediately,
            and you will want to call .close() before exiting to restore it to normal mode.
            Suggested to call .close() in high-level try:finally: so that .close() is called
            on uncaught exception.  See example at bottom of this file.
        """
        self.parent = parent
        self.border = border
        self.scrub  = scrub

        if parent is None:
            self.closed       = False
            self.run_pool     = None
            self.keystrokes   = CloseableQueue()
            self.resizes      = CloseableQueue()
            self.resize_event = Event() # Internal

            os.environ.setdefault('ESCDELAY', '25') # This makes the escape key less sluggish.

            self.c = curses.initscr()

            self.color = color
            if color:
                curses.start_color()
                self.colorPairs = {}    # Maps (fg, bg) to color index

            x = y  = 0
            height, width = self.c.getmaxyx()
            self.resizes.push((width, height))

            self.c.keypad(1)
            curses.curs_set(0)
            curses.noecho()
            curses.cbreak()

            signal.signal(signal.SIGWINCH, self.sigwinch)

        else:
            self.resize_event = None
            self.c = curses.newwin(height, width, parent.y + y, parent.x + x)
            self.color = parent.color

        if border:
            self.c.border()
            x      += 1
            y      += 1
            width  -= 2
            height -= 2

        self.x      = x
        self.y      = y
        self.width  = width
        self.height = height

        self.c.leaveok(1)   # Leave cursor wherever it lay after updates (not visible anyway so who cares).
        self.c.timeout(0)   # Non-blocking read.

    async def run(self):
        """You must spawn this for the top level Window only.
        """
        assert self.parent is None

        async with pool() as self.run_pool:
            self.run_pool.start_soon(self.monitor_stdin)
            self.run_pool.start_soon(self.monitor_resize)

    def colorPairAttribute(self, fg, bg=curses.COLOR_BLACK):
        """Pass a foreground and background color (see constants up top) and this returns
            an attribute you can logical-or in to attrs for show().
        """
        if not self.color:
            raise Exception("Must pass color=True to root window in order to use color.")
        cp = (fg, bg)
        if cp not in self.colorPairs:
            cn = len(self.colorPairs)+1
            if cn >= curses.COLOR_PAIRS:
                cn = 0
            else:
                curses.init_pair(cn, fg, bg)
            attr = curses.color_pair(cn)
            self.colorPairs[cp] = attr
        return self.colorPairs[cp]

    def show(self, x, y, text, attrs=0):
        """Paint string (or single char given as ascii integer) text at x, y with optional attributes.
        Here x, y are in the *internal* coordinates of the window, which includes inset by border if any.
        Must call flush() to actually show results.
        """

        if self.border:
            x += 1
            y += 1

        if self.scrub:
            text = self.weirdChars.sub(' ', text)

        try:
            if isinstance(text, int):
                self.c.addchr(y, x, text, attrs)
            else:
                self.c.addstr(y, x, text, attrs)
        except:
            pass    # Curses throws errors when writing to the last char on the screen, but still works...

    def clear(self):
        "Clear the contents of the window."
        self.c.erase()

    def vline(self, x, y, h, ch):
        "Display a vertical line starting at (x, y) with height of h consisting of the character ch."
        self.c.vline(y, x, ch, h)

    def flush(self):
        """Update the screen to reflect changes since last flush.
        """
        self.c.refresh()

    def close(self):
        """Restore screen/terminal to normal mode.
        Recommend calling close in try:finally: -- see example below.
        """
        if self.parent is None and not self.closed:
            self.closed = True
            if self.run_pool is not None:
                self.run_pool.cancel_scope.cancel()
            self.keystrokes.close()
            self.resizes.close()
            signal.signal(signal.SIGWINCH, signal.SIG_IGN)
            curses.nocbreak()
            curses.echo()
            curses.curs_set(1)
            self.c.keypad(0)
            # All of the above are *probably* optional, since this should do them all:
            curses.endwin()

    # ----------- Internal -------------

    async def monitor_stdin(self):

        while not self.closed:
            await wait_readable(sys.stdin)
            self.stdinAvailable()

    async def monitor_resize(self):

        while not self.closed:
            await self.resize_event.wait()
            self.resize_event = Event()
            self.stdinAvailable()

    def stdinAvailable(self):
        """Internal function.  Called whenever there is data available at stdin.
        """
        while not self.closed:

            #try:
            #    key = self.c.getkey()
            #except:
            #    return

            key = self.c.getch()
            if key == -1:
                return

            if key == curses.KEY_RESIZE:
                self.resizeEvent()
                continue

            self.keystrokes.push(key)

    def resizeEvent(self):
        """Internal function.  Called via Io/stdinAvailable when curses reports a resize.
        """
        assert self.parent is None  # For now only supported on root window.
        height, width = self.c.getmaxyx()
        if self.border:
            width  -= 2
            height -= 2
        self.width  = width
        self.height = height

        self.resizes.push((width, height))

    def sigwinch(self, signo, frame):
        """When we receive a SIGWINCH, pass it on to curses via the old handler, but
        also queue stdinAvailable() since curses will be giving us side-channel info
        which isn't actually available on stdin...
        """
        w, h = terminal_size()
        curses.resizeterm(h, w)    # Should call curses default SIGWINCH handler, but bug makes it invisible.
        self.resize_event.set()

def terminal_size():
    import fcntl, termios, struct
    h, w, hp, wp = struct.unpack('HHHH',
        fcntl.ioctl(0, termios.TIOCGWINSZ,
        struct.pack('HHHH', 0, 0, 0, 0)))
    return w, h

if __name__ == "__main__":

    import trio

    async def handle_input(win):

        async for key in win.keystrokes:
            win.show(0, 30, "[%3s]"%(key,))

            if key == ord('q'):
                win.close()

    async def handle_resize(win):

        async for width, height in win.resizes:
            win.show(0, 31, "Size: %dx%d      "%(width, height))

    async def test():

        win = Window()

        try:

            def centerX(win, y, text, attrs=0):
                win.show((win.width-len(text))//2, y, text, attrs)

            centerX(win, 20, "Hello World!")
            centerX(win, 21, " (Really!)", win.Bold)
            win.flush()

            w = Window(win, 20, 22, 30, 10, True)
            centerX(w, 0, "Test...")
            w.flush()

            z = Window(w, 0, 1, 5, 5, True)
            z.show(1, 1, "X", win.Blink)
            z.flush()

            async with pool() as p:
                p.start_soon(win.run)
                p.start_soon(handle_input , win)
                p.start_soon(handle_resize, win)
        finally:
            win.close()
            print("CLOSED CLEANLY")

    trio.run(test)

