
#
# Trax (network activity tracker) Version 1.3
#
# Copyright 2023 Brandyn Webb
# 
# This file is part of Trax (network activity tracker).
# 
# Trax is free software: you can redistribute it and/or modify it under the terms 
# of the GNU General Public License as published by the Free Software Foundation, 
# either version 3 of the License, or (at your option) any later version.
# 
# Trax is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with 
# Trax. If not, see <https://www.gnu.org/licenses/>. 
# 

"""This provides various useful trio utilities.

    - Event2, Event3, Holder, and Queue
        are trio Event-like objects that facilitate passing values between threads.

    - spawn, fork, first_to_finish, and waitfor
        allow you to pre-call your async methods and then
            start them in a pool (spawn), run them in
            parallel (fork), race them (first_to_finish),
            or escape them back into an async callable (waitfor).

    - And other stuff.
"""

import trio
import trio.testing
from trio import Event, sleep, open_nursery as pool, move_on_after, run_process

async def sim_step_complete():
    """This is how to let the current "simulation" step finish.
    This method only returns once all other tasks are blocked.
    """
    await trio.testing.wait_all_tasks_blocked()
    # Todo:
    # Might be able to use MockClock instead, to avoid
    #   heavy i/o preventing advancement: https://trio.readthedocs.io/en/stable/reference-testing.html

never_set = Event() # Don't ever set this

wait_forever = never_set.wait   # Function that never returns

class Event2(object):
    """A trio Event that allows set() to take a parameter
        which is returned by wait().

    Event2 returns first value set.
    Event3 returns last value set.
    """
    __slots__ = ('value', 'event')
    def __init__(self):
        self.event = Event()
        self.value = None

    def set(self, value=None):
        """Only the first non-None value is remembered.
        """
        if self.value is None:
            self.value = value
        self.event.set()

    async def wait(self):
        """Returns the value of the *first* (non-None) call to set()
        (Or last if an Event3 or Events)
        """
        await self.event.wait()
        return self.value

class Event3(Event2):
    """Like Event2 but returns last value set.
    """
    def set(self, value=None):
        self.value = value
        self.event.set()

class Holder(Event3):
    """This is more like a one-item, overwritable Queue than an Event.

    Made for many writers, one reader.

    This clears itself each time wait() is called, so that
        you can immediately call wait() again.

    As such, you can use it as an iterator over values passed
        via set(), with the caveat (or feature) that if many
        sets are called between waits, only the last value will
        be returned.
    """
    async def wait(self):
        """Only one task should call this at a time!
        """
        await self.event.wait()
        self.event = Event()
        return self.value

    def __aiter__(self):
        return self
    __anext__ = wait

class Queue(object):
    """Just a queue with sync push and async pop.

    Default pop implements FIFO, but pop(-1) will
        do LIFO.

    Only one task should be waiting on pop() at a time,
        but pushes can come from all over.

    Can be used as an async iterator, as in:

        async for item in queue:
            ...

    Can push from another thread, if you set threaded=True
        on creation and use push_from_thread.  (If threaded
        is True, you can only create the Queue from within
        a running trio thread.)
    """
    def __init__(self, threaded=False):
        self.q     = []
        self.event = Event()

        if threaded:
            self.trio_token = trio.lowlevel.current_trio_token()

    def push(self, item):
        self.q.append(item)
        self.event.set()

    def push_from_thread(self, item):
        """Like push but for calling from a thread.
        It should also be safe to use this from the main thread, but
            it incurs a lot of extra overhead so use ordinary push()
            whenever possible.
        """
        self.trio_token.run_sync_soon(self.push, item)
        #trio.from_thread.run_sync(self.push, item, trio_token = self.trio_token)

    async def pop(self, i=0):
        """Only one task should use this at a time.

        Multiple tasks calling this at the same time will
            hang all but one.
        """
        while not self.q:
            self.event = Event()   # Reset the event to non-triggered state.
            await self.event.wait()
        return self.q.pop(i)

    def clear(self):
        self.q = []

    def __len__(self):
        return len(self.q)

    def __getitem__(self, i):
        return self.q[i]

    def __aiter__(self):
        return self
    __anext__ = pop

class CloseableQueue(Queue):
    """Just like Queue but can be closed, which results in
        EOFError in pop or aiter() ending, once the queue
        runs empty.
    """
    def __init__(self, threaded=False):
        Queue.__init__(self, threaded)

        self.closed = False

    def push(self, item):
        if not self.closed:
            Queue.push(self, item)

    async def pop(self, i=0):
        while not self.q:
            if self.closed:
                raise EOFError
            self.event = Event()   # Reset the event to non-triggered state.
            await self.event.wait()
        return self.q.pop(i)

    async def __aiter__(self):
        while True:
            while not self.q:
                if self.closed:
                    return
                self.event = Event()
                await self.event.wait()
            yield self.q.pop(0)

    def close(self):
        self.closed = True
        self.event.set()

async def waitfor(awaitable):
    """This function is useful to escape the function call in
        something expecting an async function.

    In particular, instead of, say:

        p.start_soon(func, arg1, arg2, arg3)

    You can say:

        p.start_soon(waitfor, func(arg1, arg2, foo=arg3, bar=arg4, ...))

    Or just:

        spawn(p, func(arg1, arg2, foo=arg3, bar=arg4, ...))

    """
    return await awaitable

def spawn(p, awaitable):
    """Starts awaitable in pool p.  E.g.:

        spawn(p, func(arg1, arg2, foo=arg3, bar=arg4, ...))

    See also waitfor()
    """
    p.start_soon(waitfor, awaitable)

async def fork(*awaitables):
    """This waits for all awaitables in parallel.

    Optionally accepts a list or tuple as only argument, so you
        can call this with:

        await fork(foo(1), foo(2))

    or:
        await fork([foo(1), foo(2)])
    """
    if len(awaitables) == 1 and isinstance(awaitables[0], (list, tuple)):
        awaitables = awaitables[0]

    async with pool() as p:
        for w in awaitables:
            p.start_soon(waitfor, w)

async def first_to_finish(*awaitables):
    """This returns the return value of the first awaitable to complete.

    awaitables - one or more awaitable objects (typically
        async functions which have already been "called" with
        their parameters but not yet awaited).
    """
    if not awaitables:
        return None

    return_value = None

    async def run(awaitable):
        nonlocal return_value
        return_value = await awaitable
        p.cancel_scope.cancel()

    async with pool() as p:
        for awaitable in awaitables:
            p.start_soon(run, awaitable)

    return return_value


async def try_until(func, event, delay=2, max_delay=None, more_delay=None, timeout=None):
    """This await's func(), then waits up to delay seconds for event to
        be set.  If it's not set by then, it calls func() again,
        delays longer, and so on, until event is set.

        Subsequent delays are equal to the previous delay plus more_delay
        (so more_delay longer each time), up to max_delay.

        If more_delay is not provided, it is assumed to equal the initial delay.

    Does not return until event is set, or if func() throws an exception,
        unless timeout is provided in which case it will return True
        after approximately that long.  For timeout-induced returns,
        the delay will never be less than the original delay.
    """
    min_delay = delay       # Just for lower limit on timeouts

    if more_delay is None:
        more_delay = delay

    total_delay = 0

    while True:
        await func()
        with trio.move_on_after(delay):
            await event.wait()
            return

        total_delay += delay
        delay       += more_delay

        if max_delay and delay > max_delay:
            delay = max_delay

        if timeout:
            if total_delay >= timeout:
                return True
            if total_delay + delay > timeout:
                delay = max(min_delay, timeout - total_delay)


async def send_eventually(sock, msg, dest, logger=None):
    """This is just sock.sendto(msg, dest) except that
        it will keep retrying on error until the message goes
        out successfully.

    Logger, if provided, is called with an error message to
        log any failures.
    """
    fallback = 3   # 3 seconds first sleep
    while True:
        try:
            return await sock.sendto(msg, dest)
        except Exception as e:
            if logger is not None:
                logger("# Sendto error (sleep %s): %s"%(fallback, e))
            await sleep(fallback)
            fallback = min(fallback*2, 60)

async def recv_eventually(sock, bufsize, logger=None):
    """This is just sock.recvfrom(msg, dest) except that
        it will keep retrying on error until a message is
        received successfully.

    Logger, if provided, is called with an error message to
        log any failures.
    """
    fallback = 3   # 3 seconds first sleep

    while True:
        try:
            return await sock.recvfrom(bufsize)
        except Exception as e:
            if logger is not None:
                logger("# Recv error (sleep %s): %s"%(fallback, e))
            await sleep(fallback)
            fallback = min(fallback*2, 60)


if __name__ == "__main__":

    async def test1(x):
        print(f"A: {x}")
        await sleep(1)
        print(f"B: {x}")

    async def test2():

        async def foo(delay):
            await sleep(delay)
            return f"Waited {delay} seconds."

        print(await first_to_finish(foo(5), foo(3), foo(7)))

    async def run():
        await fork(test1(1), test1(2), test1(3))
        await test2()

    trio.run(run)

